unit FunctiiNumerologie;

interface

uses Math, DateUtils, Classes;


type
  TStagesReduce =  array[0..10] of Integer;
  TCharSet = set of Char;
  TFrequency = array[1..9] of Integer;
  TSquareCoord = array[1..3] of Integer; // coodonatele ptr square of x

const
  cst = ['a'..'z'];
  cst_validchars : TCharSet = ['a'..'z']+['A'..'Z']+['0'..'9']+[' '];
  cst_letters : TCharSet = ['a'..'z']+['A'..'Z'];
  cst_lettersLow : TCharSet = ['a'..'z'];
  cst_lettersHigh : TCharSet = ['A'..'Z'];
  cst_digits : TCharSet = ['0'..'9'];
  cst_vowels : TCharSet = ['a','e','i','o','u','A','E','I','O','U'];
  cst_consonants : TCharSet = (['a'..'z']+['A'..'Z'])-['a','e','i','o','u','A','E','I','O','U'];
  SYS_PYTHAGORA = 1;
  SYS_CHALDEAN = 2;
  SYS_INDIAN = 3;

var
  pythagora,chaldean,indian : array[Byte] of Byte;


procedure setPythagora;
procedure setChaldean;
procedure setIndian;


function reduceNrOnce(nr : Integer) : Byte;
function reduceNr(nr : Integer; stages : Boolean=False) : TStagesReduce;
function reduceDate(aDate : TDateTime; dict : Boolean=False) : TStagesReduce;
function reduceText(aText : string;stages : Boolean=False; whatsystem : Integer=SYS_PYTHAGORA) : TStagesReduce;

function validChars(aText : String) : String;
function digits(aText : String; whatsystem : Integer=SYS_PYTHAGORA) : string;
function evolve(aText : string; whatsystem : Integer=SYS_PYTHAGORA; validation : Boolean=True) : string;

function getWords(aText : string; validation : Boolean=True) : TStringList;
function getInitials(aText : string) : string;
function getLasts(aText : string) : string;
function getVowels(aText : string) : string;
function getConsonants(aText : string) : string;

function freqLetters(aText : string; whatsystem : Integer=SYS_PYTHAGORA) : TFrequency;
function freqInitials(aText : string; whatsystem : Integer=SYS_PYTHAGORA) : TFrequency;
function freqWords(aText : string; whatsystem : Integer=SYS_PYTHAGORA) : TFrequency;
function freqEvolution(aText : string; whatsystem : Integer=SYS_PYTHAGORA) : TFrequency;

function squareCoord(aNumber : Integer; aLenght : Integer) :  TSquareCoord;


implementation

uses SysUtils;

function reduceNrOnce(nr : Integer) : Byte;
begin
  { returns the digit sum of an integer }
  Result := 0;
  while nr > 0 do
  begin
    Result := Result + nr mod 10;
    nr := nr div 10;
  end;
end;

function reduceNr(nr : Integer; stages : Boolean=False) : TStagesReduce;
var i : Integer;
begin
   {
   receives a number and reduce it to a digit
   if stages=True, then it returns all the stages of reduction process. Last number in the list is the final result
    }
   for i := 0 to 10 do
    Result[i] := -1; 
   if stages then begin
     i := 1;
     Result[i] := nr;
     while nr > 9 do begin
       nr := reduceNrOnce(nr);
       i := i + 1;       
       Result[i] := nr;
     end;
   end
   else begin
     i := 1;
     while nr > 9 do
     begin
       nr := reduceNrOnce(nr);
     end;
     Result[i] := nr;
   end;

end;

function reduceDate(aDate : TDateTime; dict : Boolean=False) : TStagesReduce;
var aReduce : TStagesReduce;
    i : Integer;
begin
  { reduces a date. }
   for i := 0 to 10 do
    Result[i] := -1;   
  if dict then
  begin
    // day
    aReduce :=  reduceNr(DayOf(aDate));
    Result[1] := aReduce[1];
    // month
    aReduce :=  reduceNr(MonthOf(aDate));
    Result[2] := aReduce[1];
    // year
    aReduce :=  reduceNr(YearOf(aDate));
    Result[3] := aReduce[1];
    // day + year
    aReduce :=  reduceNr(Result[1]+Result[3]);
    Result[4] := aReduce[1];
    // day + month
    aReduce := reduceNr(Result[1]+Result[2]);
    Result[5] := aReduce[1];
    // moth + year
    aReduce := reduceNr(Result[2]+Result[3]);
    Result[6] := aReduce[1];
    // total
    aReduce := reduceNr(DayOf(aDate)+MonthOf(aDate)+YearOf(aDate));
    Result[7] := aReduce[1];
  end
  else begin
    // total
    aReduce := reduceNr(DayOf(aDate)+MonthOf(aDate)+YearOf(aDate));
    Result[1] := aReduce[1];
  end;
end;

function reduceText(aText : string;stages : Boolean=False; whatsystem : Integer=SYS_PYTHAGORA) : TStagesReduce;
var aSum,i : Integer;
    aChar : Char;
begin
  { Calculeaza suma caracterelor din textul primit
   arguments: text = string,
              whatsystem = numerological system
              stages = if True, returns list with all stages of calculation
   }
  for i := 0 to 10 do
    Result[i] := -1;
    
  aSum := 0;
  case whatsystem of
    SYS_PYTHAGORA :
      for i := 1 to Length(aText) do begin
        aChar := aText[i];
        aSum := aSum + pythagora[Ord(aChar)];
      end;
    SYS_CHALDEAN :
      for i := 1 to Length(aText) do begin
        aChar := aText[i];
        aSum := aSum + chaldean[Ord(aChar)];
      end;
    SYS_INDIAN :
      for i := 1 to Length(aText) do begin
        aChar := aText[i];
        aSum := aSum + indian[Ord(aChar)];
      end;    
  end;

  Result := reduceNr(aSum,stages);
end;

function validChars(aText : String) : String;
var aChar : Char;
    i : Integer;
begin
 { Keeps only the characters in chars string. Removes everything else }
  Result := aText;
  for i := 1 to Length(aText) do begin
    aChar := aText[i];
    if not(aChar in cst_validchars) then
      Result := StringReplace(Result,aChar,'',[rfReplaceAll]);
  end;
end;

function digits(aText : String; whatsystem : Integer=SYS_PYTHAGORA) : string;
var aChar : Char;
    i : Integer;
begin
   { Returns a string where letters are replaced with their digits }
   Result := '';
   case  whatsystem of
    SYS_PYTHAGORA :
    for i := 1 to Length(aText) do begin
      aChar := aText[i];
      Result := Result + IntToStr(pythagora[Ord(aChar)]);
    end;
    SYS_CHALDEAN :
    for i := 1 to Length(aText) do begin
      aChar := aText[i];
      Result := Result + IntToStr(chaldean[Ord(aChar)]);
    end;
    SYS_INDIAN :
    for i := 1 to Length(aText) do begin
      aChar := aText[i];
      Result := Result + IntToStr(indian[Ord(aChar)]);
    end;
   end;
end;

function evolve(aText : string;whatsystem : Integer=SYS_PYTHAGORA; validation : Boolean=True) : string;
var aSum, i : Integer;
    aChar : Char;
begin
   {  evolves the string
   la pozitia unei litere, se pune suma intermediara calculata cu litera inclusiv
   res[n] = reduceText(s[:n+1])
   example Ionescu Mihai => 9627825 99899 }
   if validation then
    aText :=  validChars(aText);
   aSum := 0;
   Result := '';
   aChar := aText[1];
   case whatsystem of
     SYS_PYTHAGORA : begin
       Result := Result + IntToStr(pythagora[Ord(aChar)]);
       aSum :=  pythagora[Ord(aChar)];
       for i := 2 to Length(aText) do
       begin
        aChar := aText[i];
        if aChar in cst_validchars then
          if aChar <> ' ' then begin
            aSum := reduceNr(aSum+pythagora[Ord(aChar)])[1];
            Result := Result + IntToStr(aSum);            
          end
          else
            Result := Result + aChar;
       end;
     end;
     SYS_CHALDEAN :  begin
       Result := Result + IntToStr(chaldean[Ord(aChar)]);
       for i := 2 to Length(aText)-1 do
       begin
        aChar := aText[i];
        if aChar in cst_validchars then
          if aChar <> ' ' then begin
            aSum := reduceNr(aSum+chaldean[Ord(aChar)])[1];
            Result := Result + IntToStr(aSum);
          end
          else
            Result := Result + aChar;
       end;
     end;
     SYS_INDIAN : begin
       Result := Result + IntToStr(indian[Ord(aChar)]);
       for i := 2 to Length(aText)-1 do
       begin
        aChar := aText[i];
        if aChar in cst_validchars then
          if aChar <> ' ' then begin
            aSum := reduceNr(aSum+indian[Ord(aChar)])[1];
            Result := Result + IntToStr(aSum);
          end
          else
            Result := Result + aChar;
       end;
     end;
   end
end;

function getWords(aText : string; validation : Boolean=True) : TStringList;
begin
   {returns a list with all the words
   if validation=True : elimina toate caracterele irelevante intai}
  Result :=  TStringList.Create;
  if validation then
    aText :=  validChars(aText);
  Result.Delimiter := ' ';
  Result.DelimitedText := aText;
end;

function getInitials(aText : string) : string;
var aList : TStringList;
    i : Integer;
begin
  { returns a string with all initials of the words of text }
  Result := '';
  aList :=  getWords(aText,True);
  for i:= 0 to aList.Count-1 do begin
    Result := Result + aList[i][1];
  end;
end;

function getLasts(aText : string) : string;
var aList : TStringList;
    i : Integer;
begin
  { returns a string with all last letter of the words of text }
  Result := '';
  aList :=  getWords(aText,True);
  for i:= 0 to aList.Count-1 do begin
    Result := Result + aList[i][Length(aList[i])];
  end;
end;

function getVowels(aText : string) : string;
var i : Integer;
begin
  { returns a string with all vowels in text }
  Result := '';
  aText :=  validChars(aText);
  for i := 1 to Length(aText) do begin
    if aText[i] in cst_vowels then
{     if i = 1 then
      Result :=  aText[i]
     else
      Result := Result + ' ' + aText[i];
}
     Result := Result + aText[i];
  end;
end;

function getConsonants(aText : string) : string;
var i : Integer;
begin
  { returns a string with all consonants in text }
  Result := '';
  aText :=  validChars(aText);
  for i := 1 to Length(aText) do begin
    if aText[i] in cst_consonants then
{     if i = 1 then
      Result :=  aText[i]
     else
      Result := Result + ' ' + aText[i];
}
    Result := Result + aText[i];
  end;
end;

function freqLetters(aText : string; whatsystem : Integer=SYS_PYTHAGORA) : TFrequency;
var i,j : Integer;
    aString : string;
begin
  { First replaces letters with digits (given the desired system)
   then returns a dictionary with frequency of digits
   example: [1:8, 2:5, 3:1, 4:2, 5:0, 6:0, 7:3, 8:2, 9:4] }

   aString :=  digits(aText,whatsystem);
   for i := 1 to 9 do begin
    Result[i] := 0;
    for j := 1 to Length(aString) do
      if StrToInt(aString[j]) = i then
         Result[i] := Result[i] + 1;
   end;
end;

function freqInitials(aText : string; whatsystem : Integer=SYS_PYTHAGORA) : TFrequency;
var i,j : Integer;
    aString : string;
begin
   { Exactly the same as freqLetters, only that it works with initials only
   gets the initials, replaces them with digits, returns dictionary with frequency of digits }
   aString :=  digits(getInitials(aText),whatsystem);   
   for i := 1 to 9 do begin
    Result[i] := 0;
    for j := 1 to Length(aString) do
      if StrToInt(aString[j]) = i then
         Result[i] := Result[i] + 1;
   end;
end;

function freqWords(aText : string; whatsystem : Integer=SYS_PYTHAGORA) : TFrequency;
var aList : TStringList;
    i,j : Integer;
    aString : string;
begin
   { same idea as freqLetters
   turns each word into a digit
   then returns a dictionary with frequency of digits }
   aList := getWords(aText,True);
   aString := '';
   for i := 0 to aList.Count-1 do begin
      aString := aString + IntToStr(reduceText(aList[i],False,whatsystem)[1]);
   end;
   for i := 1 to 9 do begin
    Result[i] := 0;
    for j := 1 to Length(aString) do
      if StrToInt(aString[j]) = i then
         Result[i] := Result[i] + 1;
   end;
end;

function freqEvolution(aText : string; whatsystem : Integer=SYS_PYTHAGORA) : TFrequency;
var i,j : Integer;
    aString : string;
begin
   { same idea as freqLetters, but combined with evolving a text
   first evolves a text => string of digits
   returns dictionary with frequency of digits }
   aString :=  evolve(aText,whatsystem,True);   
   for i := 1 to 9 do begin
    Result[i] := 0;
    for j := 1 to Length(aString) do
      if StrToInt(aString[j]) = i then
         Result[i] := Result[i] + 1;
   end;
end;

function squareCoord(aNumber : Integer; aLenght : Integer) :  TSquareCoord;
var rownr,columnnr,squarenr : Integer;
begin
  {
   Calculates coordinates (row,column,square) of nr in a square of lengthxlength dimension
   Ex: 143 in 12x12 has the coordinates (11,12,1)
   145 in 12x12 has coordinates (1,1,2)
   }
   Result[1] := 0;
   Result[2] := 0;
   Result[3] := 0;
   rownr := ((aNumber-1) mod aLenght) + 1;
   columnnr := ((aNumber div aLenght - 1)  mod aLenght) + 1;
   squarenr := (((aNumber div aLenght) div aLenght - 1) mod aLenght) + 1;
   Result[1] := rownr;
   Result[2] := columnnr;
   Result[3] := squarenr;
end;


procedure setPythagora;
begin
  pythagora[Ord('a')] := 1;
  pythagora[Ord('b')] := 2;
  pythagora[Ord('c')] := 3;
  pythagora[Ord('d')] := 4;
  pythagora[Ord('e')] := 5;
  pythagora[Ord('f')] := 6;
  pythagora[Ord('g')] := 7;
  pythagora[Ord('h')] := 8;
  pythagora[Ord('i')] := 9;
  pythagora[Ord('j')] := 1;
  pythagora[Ord('k')] := 2;
  pythagora[Ord('l')] := 3;
  pythagora[Ord('m')] := 4;
  pythagora[Ord('n')] := 5;
  pythagora[Ord('o')] := 6;
  pythagora[Ord('p')] := 7;
  pythagora[Ord('q')] := 8;
  pythagora[Ord('r')] := 9;
  pythagora[Ord('s')] := 1;
  pythagora[Ord('t')] := 2;
  pythagora[Ord('u')] := 3;
  pythagora[Ord('v')] := 4;
  pythagora[Ord('w')] := 5;
  pythagora[Ord('x')] := 6;
  pythagora[Ord('y')] := 7;
  pythagora[Ord('z')] := 8;
  pythagora[Ord('A')] := 1;
  pythagora[Ord('B')] := 2;
  pythagora[Ord('C')] := 3;
  pythagora[Ord('D')] := 4;
  pythagora[Ord('E')] := 5;
  pythagora[Ord('F')] := 6;
  pythagora[Ord('G')] := 7;
  pythagora[Ord('H')] := 8;
  pythagora[Ord('I')] := 9;
  pythagora[Ord('J')] := 1;
  pythagora[Ord('K')] := 2;
  pythagora[Ord('L')] := 3;
  pythagora[Ord('M')] := 4;
  pythagora[Ord('N')] := 5;
  pythagora[Ord('O')] := 6;
  pythagora[Ord('P')] := 7;
  pythagora[Ord('Q')] := 8;
  pythagora[Ord('R')] := 9;
  pythagora[Ord('S')] := 1;
  pythagora[Ord('T')] := 2;
  pythagora[Ord('U')] := 3;
  pythagora[Ord('V')] := 4;
  pythagora[Ord('W')] := 5;
  pythagora[Ord('X')] := 6;
  pythagora[Ord('Y')] := 7;
  pythagora[Ord('Z')] := 8;
  pythagora[Ord('1')] := 1;
  pythagora[Ord('2')] := 2;
  pythagora[Ord('3')] := 3;
  pythagora[Ord('4')] := 4;
  pythagora[Ord('5')] := 5;
  pythagora[Ord('6')] := 6;
  pythagora[Ord('7')] := 7;
  pythagora[Ord('8')] := 8;
  pythagora[Ord('9')] := 9;
  pythagora[Ord('0')] := 0;
  pythagora[Ord(' ')] := 0;  
end;

procedure setChaldean;
begin
  chaldean[Ord(' ')] := 0;
  chaldean[Ord('0')] := 0;
  chaldean[Ord('9')] := 9;
  chaldean[Ord('A')] := 1;
  chaldean[Ord('a')] := 1;
  chaldean[Ord('I')] := 1;
  chaldean[Ord('i')] := 1;
  chaldean[Ord('J')] := 1;
  chaldean[Ord('j')] := 1;
  chaldean[Ord('Q')] := 1;
  chaldean[Ord('Y')] := 1;
  chaldean[Ord('1')] := 1;
  chaldean[Ord('B')] := 2;
  chaldean[Ord('b')] := 2;
  chaldean[Ord('K')] := 2;
  chaldean[Ord('k')] := 2;
  chaldean[Ord('R')] := 2;
  chaldean[Ord('r')] := 2;
  chaldean[Ord('2')] := 2;
  chaldean[Ord('C')] := 3;
  chaldean[Ord('c')] := 3;
  chaldean[Ord('G')] := 3;
  chaldean[Ord('g')] := 3;
  chaldean[Ord('L')] := 3;
  chaldean[Ord('l')] := 3;
  chaldean[Ord('3')] := 3;
  chaldean[Ord('D')] := 4;
  chaldean[Ord('d')] := 4;
  chaldean[Ord('M')] := 4;
  chaldean[Ord('m')] := 4;
  chaldean[Ord('T')] := 4;
  chaldean[Ord('t')] := 4;
  chaldean[Ord('4')] := 4;
  chaldean[Ord('E')] := 5;
  chaldean[Ord('e')] := 5;
  chaldean[Ord('H')] := 5;
  chaldean[Ord('h')] := 5;
  chaldean[Ord('N')] := 5;
  chaldean[Ord('n')] := 5;
  chaldean[Ord('X')] := 5;
  chaldean[Ord('x')] := 5;
  chaldean[Ord('5')] := 5;
  chaldean[Ord('U')] := 6;
  chaldean[Ord('u')] := 6;
  chaldean[Ord('V')] := 6;
  chaldean[Ord('v')] := 6;
  chaldean[Ord('W')] := 6;
  chaldean[Ord('w')] := 6;
  chaldean[Ord('6')] := 6;
  chaldean[Ord('O')] := 7;
  chaldean[Ord('o')] := 7;
  chaldean[Ord('Z')] := 7;
  chaldean[Ord('z')] := 7;
  chaldean[Ord('7')] := 7;
  chaldean[Ord('F')] := 8;
  chaldean[Ord('f')] := 8;
  chaldean[Ord('P')] := 8;
  chaldean[Ord('p')] := 8;
  chaldean[Ord('8')] := 8;
end;

procedure setIndian;
begin
  indian[Ord(' ')] := 0;
  indian[Ord('0')] := 0;
  indian[Ord('9')] := 9;
  indian[Ord('A')] := 1;
  indian[Ord('a')] := 1;
  indian[Ord('I')] := 1;
  indian[Ord('i')] := 1;
  indian[Ord('J')] := 1;
  indian[Ord('j')] := 1;
  indian[Ord('Q')] := 1;
  indian[Ord('Y')] := 1;
  indian[Ord('1')] := 1;
  indian[Ord('B')] := 2;
  indian[Ord('b')] := 2;
  indian[Ord('K')] := 2;
  indian[Ord('k')] := 2;
  indian[Ord('R')] := 2;
  indian[Ord('r')] := 2;
  indian[Ord('2')] := 2;
  indian[Ord('C')] := 3;
  indian[Ord('c')] := 3;
  indian[Ord('G')] := 3;
  indian[Ord('g')] := 3;
  indian[Ord('L')] := 3;
  indian[Ord('l')] := 3;
  indian[Ord('3')] := 3;
  indian[Ord('D')] := 4;
  indian[Ord('d')] := 4;
  indian[Ord('M')] := 4;
  indian[Ord('m')] := 4;
  indian[Ord('T')] := 4;
  indian[Ord('t')] := 4;
  indian[Ord('4')] := 4;
  indian[Ord('E')] := 5;
  indian[Ord('e')] := 5;
  indian[Ord('H')] := 8; // aici e diferenta fata de chaldean
  indian[Ord('h')] := 8;
  indian[Ord('N')] := 5;
  indian[Ord('n')] := 5;
  indian[Ord('X')] := 5;
  indian[Ord('x')] := 5;
  indian[Ord('5')] := 5;
  indian[Ord('U')] := 6;
  indian[Ord('u')] := 6;
  indian[Ord('V')] := 6;
  indian[Ord('v')] := 6;
  indian[Ord('W')] := 6;
  indian[Ord('w')] := 6;
  indian[Ord('6')] := 6;
  indian[Ord('O')] := 7;
  indian[Ord('o')] := 7;
  indian[Ord('Z')] := 7;
  indian[Ord('z')] := 7;
  indian[Ord('7')] := 7;
  indian[Ord('F')] := 8;
  indian[Ord('f')] := 8;
  indian[Ord('P')] := 8;
  indian[Ord('p')] := 8;
  indian[Ord('8')] := 8;
end;


initialization
   setPythagora;
   setChaldean;
   setIndian;

end.

